"""Static XTR HTML report generator and template for CARIAD project"""

import datetime
import json
import logging
import os
import pathlib
import re
import shutil
import time
import xml.etree.ElementTree as ET
from importlib.metadata import metadata, version

import click
from jinja2 import Environment, FileSystemLoader, select_autoescape
from markupsafe import Markup

_logger = logging.getLogger("xtr_cariad_report")


def step(name: str = ""):
    """Logging wrapper for a step and it's duration"""

    def decorator(function):
        def wraps(*args, **kwargs):
            _logger.info("Step: %s", name)
            start_time = time.time()
            function(*args, **kwargs)
            end_time = time.time()
            _logger.info(":: duration: %.3f s", end_time - start_time)

        return wraps

    return decorator


def stage(name: str = ""):
    """Logging wrapper for a stage and it's duration"""

    def decorator(function):
        def wraps(*args, **kwargs):
            _logger.info("*** Stage: %s ***", name)
            start_time = time.time()
            function(*args, **kwargs)
            end_time = time.time()
            _logger.info("*** :: duration: %.3f s ***", end_time - start_time)

        return wraps

    return decorator


class ReportGenerator:
    """Report generator class with respective build functions"""

    def __init__(self, source: pathlib.Path, destination: pathlib.Path):
        self._source = source
        self._destination = destination

        self._json_report = None

        meta = metadata("xtr_report_cariad")
        self._package = {
            "name": meta.get("Name"),
            "author": meta.get("Author"),
            "version": version("xtr_report_cariad"),
        }

        self.validate_source(source=source)

    @step("Validating source folder")
    def validate_source(self, source: pathlib.Path):
        """Validate, that the provided source folder has expected structure"""

        p = source / "Global"
        if not p.is_dir():
            raise FileNotFoundError(
                f'Provided source path "{source}" is invalid! Missing "Global" folder!'
            )

        p = p / "JsonReport.json"
        if not p.is_file():
            raise FileNotFoundError(
                f'Provided source path "{source}" is invalid! Missing "JsonReport.json" file!'
            )

        p = source / "Tests"
        if not p.is_dir():
            raise FileNotFoundError(
                f'Provided source path "{source}" is invalid! Missing "Tests" folder!'
            )

    @step("Parsing JSON report")
    def _load_json_report(self):
        json_report = json.loads(
            (self._source / "Global" / "JsonReport.json").read_text("utf-8")
        )

        self._json_report = json_report

    def _assign_identifiers(self, nodes, prefix=""):

        for idx, node in enumerate(nodes):
            identifier = f"{prefix}{idx}"

            node["depth"] = identifier.count("-")
            node["identifier"] = identifier
            if node.get("children", {}):
                self._assign_identifiers(
                    nodes=node["children"], prefix=f"{identifier}-"
                )

    @staticmethod
    def find_node(nodes: list, name: str) -> dict | None:
        """Find node by it's name in the tree"""
        for node in nodes:
            if node.get("name") == name:
                return node
            else:
                children = node.get("children")
                if children:
                    res = ReportGenerator.find_node(children, name=name)
                    if res is not None:
                        return res

    def _generate_kpm_reports(self, nodes: dict):

        for idx, node in enumerate(nodes):

            files = node.get("files")
            if files:
                reports = files.get("reports")

                if node.get("status") == "FAIL" and node.get("depth", 0) > 1:
                    kpm: pathlib.Path = (
                        self._destination
                        / "files"
                        / node.get("name")
                        / "kpm.txt"
                    )

                    details = self._json_report.get("session", {}).get(
                        "details", {}
                    )

                    content = f'; KPM information for test case "{node.get("name")}" (status: "{node.get("status")}")\n\n'

                    content += f"[test_time]\n"
                    timestamp = details.get("datetime")
                    if timestamp:
                        content += f"{timestamp}\n"
                    content += "\n"

                    # Get/set description
                    d = node.get("description", None)
                    metadata = node.get("metadata", {})
                    if not d:
                        d = metadata.get("description", None)
                    d = d if d is not None else ""
                    if not isinstance(d, str):
                        d = json.dumps(d)
                    content += f"[short_text]\n{d.strip()}\n\n"

                    content += f"[precondition]\n"
                    stages = node.get("stages")
                    if stages:
                        for stage in stages:
                            for stage_item in stage.get("stages"):
                                item_name = stage_item.get("name")
                                if not item_name.startswith(
                                    "assert"
                                ) and not item_name.startswith("expect"):
                                    content += f"- {item_name}\n"
                    content += "\n"

                    content += f"[action]\n"
                    if stages:
                        for stage in stages:
                            for stage_item in stage.get("stages"):
                                item_name = stage_item.get("name")
                                if item_name.startswith(
                                    "assert"
                                ) or item_name.startswith("expect"):
                                    if stage_item.get("status") == "FAIL":
                                        content += f"- {stage_item.get('name')}: {stage_item.get('message')}\n"
                    content += "\n"

                    content += f"[actual_result]\n"
                    if stages:
                        for stage in stages:
                            for stage_item in stage.get("stages"):
                                item_name = stage_item.get("name")
                                if item_name.startswith(
                                    "assert"
                                ) or item_name.startswith("expect"):
                                    if stage_item.get("status") == "FAIL":
                                        content += f"- {stage_item.get('actual_result')}\n"
                    content += "\n"

                    content += f"[expected_result]\n"
                    if stages:
                        for stage in stages:
                            for stage_item in stage.get("stages"):
                                item_name = stage_item.get("name")
                                if item_name.startswith(
                                    "assert"
                                ) or item_name.startswith("expect"):
                                    if stage_item.get("status") == "FAIL":
                                        content += f"- {stage_item.get('expected_result')}\n"
                    content += "\n"

                    content += f"[test_requirements]\n"
                    verfies = metadata.get("verifies", [])
                    for req in verfies:
                        content += f"- {req}\n"
                    content += "\n"

                    content += f"[environement]\n"
                    for key, val in details.items():
                        content += f"{key}={val}\n"
                    content += "\n"

                    kpm.write_text(content, encoding="utf-8")

                    reports.append(
                        {
                            "name": kpm.name,
                            "path": kpm.as_posix(),
                            "rel": os.path.relpath(
                                kpm,
                                self._destination / "files" / node.get("name"),
                            ).replace("\\", "/"),
                        }
                    )

            if node.get("children", {}):
                self._generate_kpm_reports(nodes=node.get("children"))

    def _assign_files(self, folder: pathlib.Path):

        testsuite = self._json_report.get("testsuite", [])

        all_files = [
            path for path in list(folder.glob("**/*")) if path.is_file()
        ]

        for entry in all_files:

            parent_name = entry.parent.name
            parent = self.find_node(testsuite, parent_name)

            if parent is None:
                parent = self.find_node(testsuite, f'{entry.parents[1].name}.{parent_name}')

            if parent is None:
                _logger.warning(
                    'Directory "%s" could not assigned to a node from JSON report! Adding files from this directory to the test suite!',
                    parent_name,
                )
                parent = testsuite[0]

            if "files" not in parent:
                parent["files"] = {"logs": [], "traces": [], "reports": []}

            loc = self._destination / "files" / parent_name
            details = {
                "name": entry.name,
                "path": entry.as_posix(),
                "rel": os.path.relpath(entry, loc).replace("\\", "/"),
            }

            if entry.suffix == ".log":
                parent["files"]["logs"].append(details)
            elif entry.suffix in [".pcap", ".pcapng", ".zst", ".asc"]:
                parent["files"]["traces"].append(details)

    def _process_log(self, parent: str, file: dict):

        LEVEL_MAP = {
            "INF": "info",
            "DBG": "debug",
            "ERR": "error",
            "TRC": "trace",
            "FTL": "fatal",
            "WRN": "warning",
        }

        source_path = pathlib.Path(file.get("path"))

        parent: pathlib.Path = (
            self._destination / "files" / parent / source_path.name
        )
        parent.mkdir(parents=True, exist_ok=True)

        regex: re.Pattern[str] = re.compile(
            r"^\[(?P<timestamp>.[^\]]*)\] \[(?P<api>.[^\]]*)\] \[(?P<level>.[^\]]*)\] (?P<message>.+)$"
        )

        type = "xtr" if source_path.name.endswith(".xtr.log") else "regular"

        if type == "xtr":
            regex = re.compile(
                r"^\[(?P<timestamp>.[^\]]*)\] \[(?P<level>.[^\]]*)\] \[(?P<api>.[^\]]*)\] (?P<message>.+)$"
            )

        previous_lvl = "info"

        def extract_info(row: str):

            nonlocal previous_lvl

            found = regex.match(row)
            if found:
                level = (
                    LEVEL_MAP[found.group("level")]
                    if type == "xtr"
                    else found.group("level")
                )
                previous_lvl = level
                return {
                    "timestamp": found.group("timestamp"),
                    "api": found.group("api"),
                    "level": level,
                    "message": found.group("message"),
                }
            else:
                return {
                    "timestamp": "",
                    "api": "",
                    "level": previous_lvl if row.strip() != "" else "",
                    "message": row,
                }

        rows = source_path.read_text(encoding="UTF-8").split("\n")
        rows = list(map(extract_info, rows))

        target = parent / "data.js"

        with target.open("w", encoding="utf-8") as f:
            f.write(f"var log_data = {json.dumps(rows)}")

        return target

    def _get_file_routes(self, nodes: list, store: list) -> list:
        """Generate a list of routes to be generated and process log files"""
        for node in nodes:
            files = node.get("files")
            if files:

                if (
                    len(files.get("logs", [])) > 0
                    or len(files.get("traces", [])) > 0
                ):
                    dst = (
                        self._destination
                        / "files"
                        / node.get("name")
                        / "index.html"
                    )
                    store.append(
                        {
                            "template": "routes/files/[folder]/index.j2.html",
                            "base": os.path.relpath(
                                self._destination,
                                dst.parent,
                            ),
                            "destination": dst,
                            "root_folder": "files",
                            "node": node,
                        }
                    )

                for log in files.get("logs"):
                    target = self._process_log(node.get("name"), log)
                    store.append(
                        {
                            "template": "routes/files/[folder]/[file]/index.j2.html",
                            "base": os.path.relpath(
                                self._destination, target.parent
                            ),
                            "destination": target.parent / "index.html",
                            "root_folder": "files",
                            "node": node,
                            "file": log,
                        }
                    )
                for trace in files.get("traces"):
                    pass
            children = node.get("children")
            if children:
                self._get_file_routes(nodes=children, store=store)

    def _generate_routes(self) -> list[dict]:

        routes = [
            {
                "template": "routes/index.j2.html",
                "base": ".",
                "destination": self._destination / "index.html",
                "root_folder": "",
            },
            {
                "template": "routes/execution/index.j2.html",
                "base": "..",
                "destination": self._destination / "execution" / "index.html",
                "root_folder": "execution",
            },
            {
                "template": "routes/files/index.j2.html",
                "base": "..",
                "destination": self._destination / "files" / "index.html",
                "root_folder": "files",
            },
        ]

        # Generate file routes

        file_routes = []
        self._get_file_routes(
            self._json_report.get("testsuite", []), store=file_routes
        )

        return [*routes, *file_routes]

    def _dump(self):

        dest = pathlib.Path() / "dump.json"
        with open(dest, "w", encoding="UTF-8") as file:
            json.dump(
                self._json_report,
                file,
                indent=4,
            )

    @stage("Generating report")
    def generate(self):
        """Trigger generation pipeline and process input data and write to destination"""

        timestamp = datetime.datetime.now().isoformat()

        self._load_json_report()
        testsuite = self._json_report.get("testsuite", [])
        if not testsuite:
            _logger.critical(
                "JSON report does not contain any testsuites! Skipping report generation!"
            )
            return
        self._assign_identifiers(testsuite)
        self._assign_files(
            self._source / "Tests",
        )

        self._destination.mkdir(exist_ok=True, parents=True)

        templates = pathlib.Path(__file__).parent / "template"
        static = pathlib.Path(__file__).parent / "static"

        loader = FileSystemLoader(templates)
        env = Environment(loader=loader, autoescape=select_autoescape())

        def include_raw(name):
            return Markup(loader.get_source(env, name)[0])

        def attribute(input, attr_name, value):
            ET.register_namespace("", "http://www.w3.org/2000/svg")
            element = ET.fromstring(input)
            element.set(attr_name, value)
            result = ET.tostring(element, encoding="unicode", method="html")
            return Markup(result)

        def inner_svg(input):
            ET.register_namespace("", "http://www.w3.org/2000/svg")
            element = ET.fromstring(input)
            output = ""
            for child in element:
                output += ET.tostring(child, encoding="unicode", method="html")
            return Markup(output)

        env.globals["include_raw"] = include_raw
        env.filters["attribute"] = attribute
        env.filters["inner_svg"] = inner_svg
        env.filters["find_node"] = ReportGenerator.find_node

        routes = self._generate_routes()
        self._generate_kpm_reports(testsuite)

        DISPLAY_NAMES = {
            "info": {
                "os": "OS",
                "python": "Python",
                "datetime": "Datetime",
                "user": "User",
                "hostname": "Hostname",
                "inputs": "Inputs",
            },
            "metadata": {
                "tc_id": "Testcase ID",
                "status": "Status",
                "authors": "Authors",
                "reviewer": "Reviewer",
                "review_status": "Review status",
                "requirement": "Requirement",
                "targets": "Targets",
                "test_type": "Test type",
                "domain": "Domain",
                "testbench_types": "Testbench types",
                "severity": "Severity",
                "technica_jira_testcase_ids": "Technica JIRA testcase IDs",
                "id": "ID",
                "template_version": "Template version",
                "verifies": "Verifies",
                "description": "Description",
                "cb_testcase_id": "CodeBeamer testcase ID",
                "cb_testsets_ids": "CodeBeamer testset IDs",
                "tags": "Tags",
                "tag": "Tag",
            },
        }

        STATUS_NAME_MAP = {
            "PASS": "passed",
            "FAIL": "failed",
            "ERROR": "blocked",
            "SKIP": "skipped",
        }

        STATUS_COLOR_MAP_FILL = {
            "PASS": "fill-emerald-400",
            "FAIL": "fill-red-500",
            "ERROR": "fill-rose-500",
            "SKIP": "fill-amber-400",
        }

        STATUS_COLOR_MAP_STROKE = {
            "PASS": "stroke-emerald-400",
            "FAIL": "stroke-red-500",
            "ERROR": "stroke-rose-500",
            "SKIP": "stroke-amber-400",
        }

        STATUS_COLOR_MAP_BG = {
            "PASS": "bg-emerald-400",
            "FAIL": "bg-red-500",
            "ERROR": "bg-rose-500",
            "SKIP": "bg-amber-400",
        }

        STATUS_COLOR_MAP_TEXT = {
            "PASS": "text-emerald-400",
            "FAIL": "text-red-500",
            "ERROR": "text-rose-500",
            "SKIP": "text-amber-400",
        }

        STATUS_COLOR_MAP_LABEL = {
            "PASS": "bg-emerald-200 text-emerald-900 dark:bg-emerald-900 dark:text-emerald-300",
            "FAIL": "bg-red-200 text-red-900 dark:bg-red-900 dark:text-red-300",
            "ERROR": "bg-rose-100 text-rose-900 dark:bg-rose-900 dark:text-rose-300",
            "SKIP": "bg-amber-100 text-amber-900 dark:bg-amber-800 dark:text-amber-300",
        }

        LOG_LEVEL_COLOR_MAP = {
            "info": "text-emerald-400",
            "warning": "text-amber-500",
            "error": "text-red-500",
        }

        LOG_LEVELS = {
            "info": "Info",
            "debug": "Debug",
            "trace": "Trace",
            "warning": "Warning",
            "error": "Error",
            "critical": "Critical",
            "fatal": "Fatal",
        }

        SETTINGS = {
            "TREE": {"COLLAPSE_LEVEL": 2, "MODE": "overview"},
            "HTML": {
                "CUSTOM_ATTRIBUTE": "logx",
                "DARK_MODE": False,
                "SUPPORT_EMAIL": "mtf.support@technica-engineering.de",
            },
        }

        for route in routes:
            t = env.get_template(route["template"])
            dest: pathlib.Path = route.get("destination")
            dest.parent.mkdir(exist_ok=True, parents=True)

            with dest.open("w", encoding="utf-8") as f:
                f.write(
                    t.render(
                        base=route["base"].replace("\\", "/"),
                        root_folder=route["root_folder"],
                        json_report=self._json_report,
                        node=route.get("node", {}),
                        file=route.get("file", {}),
                        DISPLAY_NAMES=DISPLAY_NAMES,
                        STATUS_NAME_MAP=STATUS_NAME_MAP,
                        STATUS_COLOR_MAP_LABEL=STATUS_COLOR_MAP_LABEL,
                        STATUS_COLOR_MAP_FILL=STATUS_COLOR_MAP_FILL,
                        STATUS_COLOR_MAP_STROKE=STATUS_COLOR_MAP_STROKE,
                        STATUS_COLOR_MAP_BG=STATUS_COLOR_MAP_BG,
                        STATUS_COLOR_MAP_TEXT=STATUS_COLOR_MAP_TEXT,
                        LOG_LEVEL_COLOR_MAP=LOG_LEVEL_COLOR_MAP,
                        LOG_LEVELS=LOG_LEVELS,
                        SETTINGS=SETTINGS,
                        pkg=self._package,
                        timestamp=timestamp,
                    )
                )

        # Copy static directory assets to the output
        shutil.copytree(
            static, self._destination / "static", dirs_exist_ok=True
        )


def main(source: pathlib.Path, destination: pathlib.Path):
    """Main entrypoint to generate report"""

    try:
        generator = ReportGenerator(source=source, destination=destination)
    except Exception as e:
        _logger.critical("Failed to initialize report generator!")
        _logger.critical(e)
        return

    generator.generate()


@click.command()
@click.option(
    "-s",
    "--source",
    required=True,
    type=click.Path(
        exists=True, dir_okay=True, file_okay=False, path_type=pathlib.Path
    ),
    help="Source folder of the XTR results",
)
@click.option(
    "-d",
    "--destination",
    required=True,
    type=click.Path(
        exists=False, dir_okay=True, file_okay=False, path_type=pathlib.Path
    ),
    help="Destination folder of the HTML report",
)
def cli_main(source: pathlib.Path, destination: pathlib.Path):
    main(source=source, destination=destination)
